'' parallel 20x4 lcd driver for the DracBlade including a 'Terminal' that scrolls properly, backspace etc

CON


  _clkmode      = xtal1 + pll16x                        ' use crystal x 16
  _xinfreq      = 5_000_000
  LatchDirection =%00000000_00000000_00001111_11111111 ' 138 active, gate active and 8 data lines active
  MaskLowByte =   %00000000_00000000_00000000_11111111 ' used in lots of routines to & mask off low byte                                     
  LCDLatch =      %00000000_00000000_00001011_00000000 ' LCD latch is xxxx101x and gate high xxxxxxx1
  EnableBit =     %00000000_00000000_00000000_00000010 ' Enable pin 6 on LCD displays

OBJ
 delay    : "timing"
  
VAR

long Char                       ' Character To Send To LCD
long Inst                       ' Instuction To Send To LCD
long Column                     ' column counter for line 4 ' don't use byte, causes all sorts of overrun errors
long LineNumber                 ' line number 1,2,3,4
long LCDChar                    ' character to send to LCD 0-$FF
long LCDByte                    ' byte to latch out to LCD display
byte Display[79]                ' buffer for display 20x4  ' put this last as bizarre overwriting going on
PUB Start  | n
    Init_Lcd
    Column:=0                                           ' start at zero
    repeat n from 0 to 79                ' fill the display memory with spaces
      Display[n]:=" "         'fill with spaces
    Redraw
{
    repeat 25        ' test code for running standalone
      out(65)

    out(10)
    out(66)
    out(13)
}    


PUB str(stringptr)
'' Print a zero-terminated string
  repeat strsize(stringptr)
    out(byte[stringptr++])

PUB dec(value) | i '' Print a decimal number
  if value < 0
    -value
    out("-")

  i := 1_000_000_000

  repeat 10
    if value => i
      out(value / i + "0")
      value //= i
      result~~
    elseif result or i == 1
      out("0")
    i /= 10

PUB hex(value, digits) '' Print a hexadecimal number
  value <<= (8 - digits) << 2
  repeat digits
    out(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))

PUB bin(value, digits)

'' Print a binary number

  value <<= 32 - digits
  repeat digits
    out((value <-= 1) & 1 + "0")

  
PUB out(c) | i,j
' Output a character, ignores most <32 and all >126
' possibly add a 'turn on display' and 'turn off display' with selected ascii characters ?? esc seq
    i:=c ' need a local working variable as an if statement seems to corrupt subsequent working variables!
    j:=Column ' and need a local version of column otherwise corrupts this too

    if c==10                                              ' Linefeed
        Newline
        Column:=0                                           ' reset column counter
        j:=21                                               ' ensures code below doesn't run


    if j<20 ' this LCD driver code is slow so better to not print anything after line 20 and wait for CRLF
    
'    if j > 19                                      ' is it off the end of the display? (old display wrapped around)
'      Column:=0                                         ' reset to beginning
'      Newline

      if c==08                                              ' Backspace
        Backspace

'      if c==13                                              ' do nothing as LF includes CR as part of the code
  
      if i >= 31                                              ' and statements don't work in if lines?!
        if i<=126                                             ' send the ascii character
          char := c                                           ' get the character
          Send_Text                                           ' send it out
          Display[Column+60]:=c                                  ' store the character in the last column
          Column:=Column+1                                    ' add 1 to column

     if j==19
        CursorOff
 
PUB CursorOn
  Inst:=13                                              ' turn on the cursor
  Send_Inst

PUB CursorOff
  Inst:=12                                          ' cursor off
  Send_Inst                                         ' send the instruction
 
PRI  Init_Lcd
    LCDLow                                              ' set the 6 LCD lines low
    delay.pause1ms(200)                                 ' short delay for LCD bootup 200ms
    LCDByte:=%00001100                                  ' send out 0011 already shifted to xxnnnnxx
    PulsOut
    LCDByte:=%00001100                                  ' send out 0011 already shifted to xxnnnnxx
    PulsOut
    LCDByte:=%00001100                                  ' send out 0011 already shifted to xxnnnnxx
    PulsOut
    LCDByte:=%00001000                                  ' send out 0010 already shifted to xxnnnnxx
    PulsOut
    LCDByte:=%00001000                                  ' send out 0010 already shifted to xxnnnnxx
    PulsOut
    LCDByte:=%00100000                                  ' send out 1000 already shifted to xxnnnnxx
    PulsOut
    Inst:=14                                            ' cursor on
    Send_Inst
    Inst:=1                                             ' cls
    Send_Inst
    Inst:=13                                            ' flashing cursor
    Send_Inst
    Inst:=128                                           ' to position 1
    Send_Inst

pri Send_Inst ' send instruction in Inst
    RSLow                                                      ' not actually needed see below but makes code more logical                 
    LCDByte:=Inst                                              ' get high nibble
    LCDByte:=LCDByte>>2                                        ' shift right two bits so xxnnnn00
    LCDByte:=LCDByte & %00111100                               ' mask the nibble and sets RS low as well (bit0)
    Pulsout                                                    ' send it out
    LCDByte:=Inst                                              ' get low nibble
    LCDByte:=LCDByte<<2                                        ' shift left two bits so xxnnnn00
    LCDByte:=LCDByte & %00111100                               ' mask the nibble and sets RS low as well (bit0)
    Pulsout                                                    ' send it out
    RSHigh                                                     ' back to text mode
 
pri Send_Text ' pass Char = a byte
   'RS always high = bit 0, bits are xxnnnnER where nnnn is the nibble, E=enable and R=RS
   LCDByte:=Char
   LCDByte:=LCDByte>>2                                  ' shift right two bits so xxnnnn00
   LCDByte:=LCDByte & %00111100                         ' mask off just the nibble
   LCDByte:=LCDByte | %00000001                         ' keep RS high
   Pulsout                                              ' send it out                                    
   LCDByte:=Char                                        ' get low nibble
   LCDByte:=LCDByte<<2                                  ' shift left two bits so xxnnnn00
   LCDByte:=LCDByte | %00000001                         ' keep RS high
   Pulsout

pub Next_Line
  
    Inst := %11000000                         ' Move Cursor To Line 2
    Send_Inst
 
PRI ToggleGate
   OUTA[8]:=0                                            ' set gate pin low
   OUTA[8]:=1                                            ' and high again - may need a delay in PASM
PRI EnableLatchPins
    DIRA:=LatchDirection                                ' 138 active, gate active and 8 data lines active
PRI Tristate ' tristate all pins
  DIRA:=%00000000_00000000_00000000_00000000                                                                                              
PUB LCDOutput | n ' sends byte LCDByte to latch
   EnableLatchPins ' turn on the pins needed to talk to the latches
   n:=LCDByte                                           ' byte value to send out
   n:=(n & MaskLowByte)                                 ' 2 AND mask the low byte
   n:=(n | LCDLatch  )                                  ' 3 OR with the LCD address 138
   OUTA:=n                                              ' 4 send it out
   ToggleGate                                           ' 5 and latch it
   tristate ' back to tristate on all pins
PUB LCDLow ' sets all LCD pins low for startup
   LCDByte:=0
   LCDOutput
PUB PulsOut ' pulse the enable pin 6 high then low - uses LCDByte and returns it unchanged
   LCDByte:=(LCDByte | %00000010)                       ' logical OR  to set just this bit high
   LCDOutput                                            ' send it out
'   delay.pause1ms(1)                                    ' short pause not needed as spin very fast
   LCDByte:=(LCDByte & %11111101)                       ' logical AND set just this bit low
   LCDOutput                                            ' send it out
'   delay.pause1ms(1)                                    ' delay for LCD to process it
PUB RSHigh  ' set RS pin 4 high - uses LCDByte and returns it unchanged
   LCDByte:=(LCDByte | %00000001)                       ' logical OR  to set just this bit high
   LCDOutput  
PUB RSLow ' set RS pin 4 low - uses LCDByte and returns it unchanged
   LCDByte:=(LCDByte & %11111110)                       ' logical AND set just this bit low
   LCDOutput  

PUB Backspace | i,j 
' backspace is a special case - need to subtract 1 off the column, make the last character a space
' and redraw the screen and print out column number of characters on last line
' only really works on line 4 but most of the time the display is on line 4 anyway
    j:=Column                 ' the last character added 1 to column so start with one less
    if j>1                      ' not reliable at 1
      Display[Column+60-1]:=32    ' was one past the character
      Column:=Column-1            ' subtract 1
      Redraw
      repeat i from 0 to j-2
        Char:=Display[60+i]        ' print out the last row with one less character
        Send_Text                     


PUB Redraw                        ' redraw the entire screen
    Inst:=128                     ' line 1
    Send_Inst
    LineNumber:=1                 ' print line 1
    Printline
    Inst:=192                     ' line 2
    Send_Inst
    LineNumber:=2                 ' print line 2
    Printline
    Inst:=148                     ' line 3
    Send_Inst
    LineNumber:=3                 ' print line 3
    Printline
    Inst:=212                     ' line 4
    Send_Inst
    LineNumber:=4                 ' print line 4
    Printline
    Inst:=212                     ' cursor to beginning of line 4
    Send_Inst
    CursorOn
PUB PrintLine | n                   ' Prints the line in LineNumber 1 to 4
    n:=(LineNumber-1)*20                ' convert to pointer for the array 0-79
    repeat 20                       ' do this 20 times
      Char:=Display[n]              ' get the character from the array
      Send_Text                     ' send it out
      n:=n+1                        ' increment counter             

PUB Newline | i ' move the lines all up one, clear line 4, move cursor to position 1 of line 4
    repeat i from 0 to 59                                            ' move up to next line
      Display[i]:=Display[i+20]

    repeat i from 60 to 79
      Display[i]:=32                                    ' space

    Column:=0                                           ' column counter to zero
    Redraw                                              ' redraw the screen
        
